@@ -29,7 +29,7 @@
 #include "file-utils.h"
 #include "error.h"                      // verbosity_level
 #include "link-includes.h"
-#include "utilities.h"                  // lg_strerror_r
+#include "utilities.h"                  // lg_strerror
 
 #ifdef _WIN32
 	#define DIR_SEPARATOR "\\"
@@ -441,6 +441,19 @@ FILE *linkgrammar_open_data_file(const char *filename)
 
 /* ======================================================== */
 
+bool check_db(const char *lang)
+{
+	char *dbname = join_path (lang, "dict.db");
+	bool retval = file_exists(dbname);
+	free(dbname);
+#if !HAVE_SQLITE3
+	if (retval)
+		prt_error("Error: Could not open dictionary \"%s\" "
+		          "(not configured with SQLite support)\n", dbname);
+#endif /* !HAVE_SQLITE3 */
+	return retval;
+}
+
 /**
  * Check to see if a file exists.
  */
@@ -506,7 +519,7 @@ char *get_file_contents(const char * dict_name)
 			{
 				char errbuf[64];
 
-				lg_strerror_r(errno, errbuf, sizeof(errbuf));
+				lg_strerror(errno, errbuf, sizeof(errbuf));
 				fclose(fp);
 				prt_error("Error: %s: Read error (%s)\n", dict_name, errbuf);
 				free(contents);
